import datetime as dt
import os
import textwrap
from unittest import TestCase
from unittest.mock import call, patch

from click.testing import CliRunner

from loggertodb import cli


class NonExistentConfigFileTestCase(TestCase):
    def setUp(self):
        runner = CliRunner(mix_stderr=False)
        self.result = runner.invoke(cli.main, ["nonexistent.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 1)

    def test_error_message(self):
        self.assertIn(
            "No such file or directory: 'nonexistent.conf'", self.result.stderr
        )


class MissingBaseUrlTestCase(TestCase):
    def setUp(self):
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        username = a_user
                        password = a_password
                        """
                    )
                )
            self.result = runner.invoke(cli.main, ["loggertodb.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 1)

    def test_error_message(self):
        self.assertIn("No option 'base_url' in section: 'General'", self.result.stderr)


class NonExistentLogLevelTestCase(TestCase):
    def setUp(self):
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        base_url = https://example.com
                        username = a_user
                        password = a_password
                        loglevel = NONEXISTENT_LOG_LEVEL
                        """
                    )
                )
            self.result = runner.invoke(cli.main, ["loggertodb.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 1)

    def test_error_message(self):
        self.assertIn(
            "loglevel must be one of ERROR, WARNING, INFO, DEBUG", self.result.stderr
        )


class ConfigurationWithNoMeteologgersTestCase(TestCase):
    def setUp(self):
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        base_url = https://example.com
                        username = a_user
                        password = a_password
                        """
                    )
                )
            self.result = runner.invoke(cli.main, ["loggertodb.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 1)

    def test_error_message(self):
        self.assertIn("No stations have been specified", self.result.stderr)


class ConfigurationWithUnsupportedFormatTestCase(TestCase):
    @patch("loggertodb.cli.EnhydrisApiClient")
    def setUp(self, mock_client):
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        base_url = https://example.com
                        username = a_user
                        password = a_password

                        [My station]
                        storage_format = unsupported
                        """
                    )
                )
            self.result = runner.invoke(cli.main, ["loggertodb.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 1)

    def test_error_message(self):
        self.assertIn("Unsupported format 'unsupported'", self.result.stderr)


class CorrectConfigurationTestCase(TestCase):
    @patch("loggertodb.cli.EnhydrisApiClient")
    @patch("loggertodb.cli.update_database")
    @patch("loggertodb.meteologgerstorage.MeteologgerStorage_simple")
    def setUp(self, mock_meteologgerstorage, mock_update_db, mock_client):
        self.mock_meteologgerstorage = mock_meteologgerstorage
        self.mock_update_db = mock_update_db
        self.mock_client = mock_client
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        base_url = https://example.com
                        username = a_user
                        password = a_password

                        [My station]
                        storage_format = simple
                        station_id = 1334
                        path = .
                        fields = 1,2,3
                        """
                    )
                )
            self.result = runner.invoke(cli.main, ["loggertodb.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 0)

    def test_has_used_base_url(self):
        self.mock_client.assert_called_once_with("https://example.com")

    def test_has_logged_on(self):
        self.mock_client.return_value.login.assert_called_once_with(
            "a_user", "a_password"
        )

    def test_has_updated_database(self):
        self.mock_update_db.assert_called_once_with(
            self.mock_client.return_value, self.mock_meteologgerstorage.return_value
        )


class CorrectConfigurationWithLogFileTestCase(TestCase):
    @patch("loggertodb.cli.EnhydrisApiClient")
    @patch("loggertodb.cli.update_database")
    @patch("loggertodb.meteologgerstorage.MeteologgerStorage_simple")
    def test_creates_log_file(self, *args):
        self.mock_meteologgerstorage = args[0]
        self.mock_update_db = args[1]
        self.mock_client = args[2]
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        base_url = https://example.com
                        username = a_user
                        password = a_password
                        logfile = deleteme

                        [My station]
                        storage_format = simple
                        station_id = 1334
                        path = .
                        fields = 1,2,3
                        """
                    )
                )
            self.result = runner.invoke(cli.main, ["loggertodb.conf"])
            self.assertTrue(os.path.exists("deleteme"))


class UpdateDatabaseTestCase(TestCase):
    @patch("loggertodb.cli.EnhydrisApiClient")
    @patch("loggertodb.cli.meteologgerstorage.MeteologgerStorage_simple")
    @patch("loggertodb.cli.HTimeseries", new=lambda x: x)
    def setUp(self, mock_meteologgerstorage, mock_client):
        # Configure EnhydrisApiClient mock
        self.mock_client = mock_client
        attrs = {
            "return_value.get_ts_end_date.return_value": dt.datetime(2019, 3, 5, 7, 20)
        }
        self.mock_client.configure_mock(**attrs)

        # Configure MeteologgerStorage mock
        self.mock_meteologgerstorage = mock_meteologgerstorage
        attrs = {
            "return_value.timeseries_ids": {1, 2, 3},
            "return_value.get_recent_data.side_effect": [
                "new data for timeseries_id=1",
                "new data for timeseries_id=2",
                "new data for timeseries_id=3",
            ],
        }
        self.mock_meteologgerstorage.configure_mock(**attrs)

        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        base_url = https://example.com
                        username = a_user
                        password = a_password

                        [My station]
                        storage_format = simple
                        station_id = 1334
                        path = .
                        fields = 1,2,3
                        """
                    )
                )
            runner.invoke(cli.main, ["loggertodb.conf"])

    def test_called_get_recent_data_as_needed(self):
        self.mock_meteologgerstorage.return_value.get_recent_data.assert_has_calls(
            [
                call(1, dt.datetime(2019, 3, 5, 7, 20)),
                call(2, dt.datetime(2019, 3, 5, 7, 20)),
                call(3, dt.datetime(2019, 3, 5, 7, 20)),
            ]
        )

    def test_called_post_tsdata_as_needed(self):
        self.mock_client.post_tsdata.has_calls(
            [
                call(1, "new data for timeseries_id=1"),
                call(2, "new data for timeseries_id=2"),
                call(3, "new data for timeseries_id=3"),
            ]
        )



import configparser
import logging
import sys
import traceback
from datetime import datetime

import click
from enhydris_api_client import EnhydrisApiClient
from htimeseries import HTimeseries

from . import __version__, meteologgerstorage
from .exceptions import LoggerToDbError


class WrongValueError(configparser.Error):
    pass


class UnsupportedFormat(Exception):
    def __str__(self):
        return "Unsupported format '{}'".format(self.args[0])


@click.command()
@click.argument("configfile")
@click.version_option(version=__version__, prog_name="loggertodb")
def main(configfile):
    """Insert meteorological logger data to Enhydris"""

    # Start by setting logger to stdout; later we will switch it according to config
    logger = logging.getLogger("loggertodb")
    stdout_handler = logging.StreamHandler()
    logger.addHandler(stdout_handler)

    try:
        config = configparser.ConfigParser(interpolation=None)
        with open(configfile) as f:
            config.read_file(f)

        # Read the [General] section
        base_url = config.get("General", "base_url")
        username = config.get("General", "username")
        password = config.get("General", "password")
        logfile = config.get("General", "logfile", fallback="")
        loglevel = config.get("General", "loglevel", fallback="warning")
        log_levels = ("ERROR", "WARNING", "INFO", "DEBUG")
        if loglevel.upper() not in log_levels:
            raise WrongValueError("loglevel must be one of " + ", ".join(log_levels))

        # Remove [General] and make sure there are more sections
        config.pop("General")
        if not len(config.sections()):
            raise configparser.NoSectionError("No stations have been specified")

        # Setup logger
        logger.setLevel(loglevel.upper())
        if logfile:
            logger.removeHandler(stdout_handler)
            logger.addHandler(logging.FileHandler(logfile))

        # Log start of execution
        logger.info("Starting loggertodb, " + datetime.today().isoformat())

        # Connect to Enhydris
        client = EnhydrisApiClient(base_url)
        client.login(username, password)

        # Read each section and do the work for it
        for section_name in config.sections():
            section = config[section_name]
            klassname = "MeteologgerStorage_" + section["storage_format"]
            if not hasattr(meteologgerstorage, klassname):
                raise UnsupportedFormat(section["storage_format"])
            klass = getattr(meteologgerstorage, klassname)
            meteologger_storage = klass(section, logger=logger)
            try:
                update_database(client, meteologger_storage)
            except LoggerToDbError as e:
                msg = "Error while processing item {}: {}".format(section, str(e))
                sys.stderr.write(msg + "\n")
                logger.error(msg)
                logger.debug(traceback.format_exc())

        # Log end of execution
        logger.info("Finished loggertodb, " + datetime.today().isoformat())
    except Exception as e:
        logger.error(str(e))
        logger.debug(traceback.format_exc())
        raise click.ClickException(str(e))


def update_database(client, meteologger_storage):
    station_id = meteologger_storage.station_id
    ts_end_dates = {
        ts_id: client.get_ts_end_date(station_id, ts_id) or datetime(1700, 1, 1)
        for ts_id in meteologger_storage.timeseries_ids
    }
    for ts_id, ts_end_date in sorted(ts_end_dates.items(), key=lambda x: x[1]):
        new_data = meteologger_storage.get_recent_data(ts_id, ts_end_date)
        if len(new_data):
            client.post_tsdata(station_id, ts_id, HTimeseries(new_data))


if __name__ == "__main__":
    sys.exit(main())
